/* Emperor - an X windows strategy game for 1-6 players
   Copyright (C) 1998 Frank Rustemeyer.
   This file is part of the Emperor package.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

   If you have changes and you want them to be included in a future
   release of Emperor, please e-mail them to mambres@mud.tap.de.

   This file implements the X windows visualisation of the game.

   98/11/30, Frank Rustemeyer <mambres@mud.tap.de>: Released.
   99/01/29, Frank Rustemeyer <mambres@mud.tap.de>: Added roads.
   99/01/29, Frank Rustemeyer <mambres@mud.tap.de>: Added language support.
   99/02/07, Frank Rustemeyer <mambres@mud.tap.de>: Added "Forces" Window.
   99/02/09, Frank Rustemeyer <mambres@mud.tap.de>: Included cities in the Forces Window
   99/04/05, Frank Rustemeyer <mambres@mud.tap.de>: Bugfix: Handle save games without units
   99/08/15, Frank Rustemeyer <mambres@mud.tap.de>: Added a counter for game turns
   99/08/15, Frank Rustemeyer <mambres@mud.tap.de>: Bugfix: mainwndw->forces wasn't initialised
   00/05/22, Frank Rustemeyer <mambres@mud.tap.de>: Ported to QT
   00/07/02, Frank Rustemeyer <mambres@mud.tap.de>: Added helicopters, bombers, carriers
   00/07/20, Frank Rustemeyer <mambres@mud.tap.de>: Workaround for a QT bug.
   04/05/22, Frank Rustemeyer <mambres@mud.tap.de>: Renamed classes: list -> llist, map -> mmap
   04/05/22, Frank Rustemeyer <mambres@mud.tap.de>: Fix for QT3 (dlg->exec() instead of dlg->show())
   04/11/27, Frank Rustemeyer <mambres@mud.tap.de>: Network support
*/

#include "emp.h"
#include "empx.h"
#include "empgame.h"

#include <qstring.h>
#include <qpainter.h>
#include <qmessagebox.h>
#include <qpixmap.h>
#include <qscrollview.h>
#include <qmenubar.h>
#include <qpopupmenu.h>
#include <qstatusbar.h>
#include <qevent.h>
#include <qpushbutton.h>
#include <qprogressbar.h>
#include <qlistbox.h>
#include <qlabel.h>
#include <qfiledialog.h>
#include <qapplication.h>
#include <qcursor.h>

#include "land.xpm"
#include "flags.xpm"
#include "units.xpm"

#include "flag1.xpm"
#include "flag2.xpm"
#include "flag3.xpm"
#include "flag4.xpm"
#include "flag5.xpm"
#include "flag6.xpm"



#include <string.h>
  // strcpy, strcat in mainwndw::invalidatemoves
//#include <stdio.h> included in empgame.h
  // sprintf in mainwndw::settitle,
#include <stdlib.h>
  // malloc() and free() in armywndw

game *g;

// **************************************************** mainwndw ****

mainwndw::mainwndw(QWidget *parent, const char *name)
      : QMainWindow(parent, name)
{   int m_height;
    setupwndw * sw;

    setCaption("Emperor");
    setMinimumSize(100, 140);

    fileopts = new QPopupMenu(this);
    menuBar()->insertItem(T_GAME, fileopts);
    fileopts->insertItem(T_SAVE, this, SLOT(file_save()));
    fileopts->insertItem(T_LOAD, this, SLOT(file_open()));
    fileopts->insertItem(T_MAP, this, SLOT(display_bird()));
    fileopts->insertItem(T_STATS, this, SLOT(display_score()));
    fileopts->insertItem(T_FORCES, this, SLOT(display_armies()));
    fileopts->insertItem(T_EXIT, this, SLOT(finish()));

    zugopts = new QPopupMenu(this);
    menuBar()->insertItem(T_MOVE, zugopts);
    zugopts->insertItem(T_DONE, this, SLOT(nextplayer()));
    zugopts->insertItem(T_MINES, this, SLOT(button_mine()));
    zugopts->insertItem(T_ROAD, this, SLOT(button_road()));
    zugopts->insertItem(T_COMPMOVE, this, SLOT(button_comp()));
    zugopts->insertItem(T_STAY, this, SLOT(button_stay()));

    nextbutton  = menuBar()->insertItem("->", this, SLOT(button_next()));
    stoppbutton = menuBar()->insertItem(T_STOP, this, SLOT(button_stopp()));
    infobutton  = menuBar()->insertItem(T_UNIT, this, SLOT(button_info()));

    canv = new canvas(this);

    resize(600, 440);
    m_height = menuBar()->heightForWidth(300);
    canv->move(0, m_height);
    m_height += statusBar()->height();
    canv->resize(600, 440 - m_height);

    qtm = new QTimer(this);
    connect(qtm, SIGNAL(timeout()), this, SLOT(heartbeat()));

    waiting = 0;
    selected = (unit *) NULL;
    selectedpos.set(-1, -1);

    g = new game();
    g->mp = new mmap();

    sw = new setupwndw(this);
       sw->exec();
    delete sw;

    g->initcities();
    
    show();
    bird = (birdwndw *) NULL;
    ctyw = (citywndw *) NULL;
    score = (scorewndw *) NULL;
    forces = (armywndw *) NULL;

    nextplayer();
    settitle();
    }

void mainwndw::busy(int b)
{   if (b) QApplication::setOverrideCursor(waitCursor);
    else QApplication::restoreOverrideCursor();
    }

void mainwndw::resizeEvent(QResizeEvent *qre)
{   int x,y;
    x = qre->size().width();
    y = qre->size().height();
    canv->resize(x, y - menuBar()->heightForWidth(x) - statusBar()->height());
    }

void mainwndw::finish()
{   if (waiting) return;
    if (bird) bird->close();
    if (ctyw) ctyw->close();
    if (score) score->close();
    if (forces) unforces();
    close();
    }

void mainwndw::enable(int button)
{   menuBar()->setItemEnabled(button, 1);
    }

void mainwndw::disable(int button)
{   menuBar()->setItemEnabled(button, 0);
    }

void mainwndw::button_stay(void)
{   if (waiting) return;
    if (!selected) return;
    selected->fortify();
    select((unit *) NULL);
    }

void mainwndw::button_mine(void)
{   if (waiting) return;
    if (!selected) return;
    if (selected->id() != MINENLEGER) 
    {   statusBar()->message(T_ERROR_NO_MINE_BOAT);
        return;
        }
    if (g->stadt(selectedpos)) 
    {   statusBar()->message(T_ERROR_MINE_IN_CITY);
        return;
        }
    g->minelegen(selectedpos);
    invalidatesquare(selectedpos);
    }

void mainwndw::button_road(void)
{   if (waiting) return;
    if (!selected) return; 
    if (selected->id() != INFANTRIE) 
    {   statusBar()->message(T_ERROR_NO_INFANTRY);
        return;
        }
    if (!(selected->restmoves())) 
    {   statusBar()->message(T_ERROR_NO_MOVEMENT);
        return;
        }
    if (!g->freiland(selectedpos))
    {   statusBar()->message(T_ERROR_NO_FREE_LAND);
        return;
        }
    (*(g->mp))[selectedpos] = 7; // make road
    selected->countmove();
    select((unit *) NULL);
    }

void mainwndw::button_comp(void)
{   if (waiting) return;
    busy(1);
    g->compmove();
    select((unit *) NULL);
    findnextunit();
    busy(0);
    invalidate(FULL);
    if (forces) forces->update_armies();
    }

void mainwndw::button_next(void)
{   if (waiting) return;
    if (!nextunit) return;
    setwindow(nextunit->getpos());
    select(nextunit);
    findnextunit();
    }

void mainwndw::button_stopp(void)
{   if (waiting) return;
    if (!selected) return;
    if (!selected->restmoves()) 
    {   statusBar()->message(T_ERROR_NO_MOVEMENT);
        return; 
        }
    selected->nomoves();
    invalidatemoves();
    if (g->stadt(selectedpos)) 
    {   char msg[120];
        statusBar()->message(selected->repair(msg));
        }
    disable(stoppbutton);
    
    //proceed: switch to next unit
//    findnextunit();
//    if (!nextunit)
//    {   disable(stoppbutton);
//       invalidatemoves();
//        }
//    button_next();
    }

void mainwndw::button_info(void)
{   if (waiting) return;
    if (!selected) return;
    if ((selected->is_transporter()) && ((transporter *) selected)->getnextunload())
    {   unloadwndw *dialog;
        dialog = new unloadwndw(this, (transporter *) selected);
        dialog->exec();
        delete dialog;
        }
    else 
    {   char s[40];
        QMessageBox::information(this, T_UNIT, selected->hitpointstring(s), "OK");
        }
    }

void mainwndw::file_open()
{   int i;  
    netwndw* nw;
    char text[60];
    QString file;

    if (waiting) return;

    file = QFileDialog::getOpenFileName(QString::null, "*.emp", this);

    if (ctyw) uncity();

    nw = new netwndw(this);
      nw->exec();
    delete nw;  
   
    select((unit *) NULL);

    g->activeplayer = 0;
    invalidate(FULL);

    if (!g->load(file)) 
    {   QMessageBox::warning(this, "Emperor", T_ERRORLOADING, "OK");
        }
    else
    {   unit *u;

        sprintf(text, T_PLAYERSTURN, g->sp[g->activeplayer].name);
        QMessageBox::information(this, "Emperor", text, "OK");
 
        u = (unit *) FIRSTMEMBER(g->units);
        if ((u) && (u != (unit *) (u->last()->first())))
        {   QMessageBox::warning(this, "Emperor", T_DATAERROR, "OK");
            }

        if (g->sp[g->activeplayer].ext)
        {   waiting = 1;
            i = g->activeplayer;
            do
               if (++i > 6) i = 1;
            while (!g->sp[i].used || g->sp[i].ext);
            g->activeplayer = i;
            busy(1);

            listener = new listen_socket(g->get_port());
            qtm->start(NETWORK_POLL_INTERVAL);
            enable(nextbutton);
            enable(stoppbutton);
            }
         else
         {   findnextunit();
             button_next();
             }
         settitle();
         g->lookout(g->activeplayer);
         invalidate(FULL);
         if (forces) forces->update();
         }
    }

void mainwndw::file_save()
{   QString file;
    if (waiting) return;
    file = QFileDialog::getSaveFileName(QString::null, "*.emp", this); 
    if (!g->save(file))
    {   QMessageBox::warning(this, "Emperor", T_ERRORSAVING, "OK");
        }
    }

mainwndw::~mainwndw(void)
{   delete canv;
    delete g;
    }

void mainwndw::invalidate(int mode)
{   canv->updateContents(0, 0, SX*20 , SY*20);
    if (bird) bird->invalidate();
    } 

void mainwndw::settitle(void)
{   char t[100];
    if (waiting)
        sprintf(t, T_WAITING);
    else if (g->activeplayer)
    {   sprintf(t, "Emperor: %s (" T_TURN " %d)", 
                   g->sp[g->activeplayer].name, g->get_turn());
	}
    else sprintf(t, "Emperor (" T_TURN " %d)", g->get_turn());
    setCaption(t);
    }

void mainwndw::invalidatesquare(coordinate c)
{   if (!g->activeplayer) return;
    if (!ONMAP(c)) return;
    canv->repaintContents(c.x * 20 -2, c.y * 20 - 2, c.x * 20 + 20, c.y * 20 + 20, FALSE);
    }

void mainwndw::invalidatearea(coordinate c1, coordinate c2)
{   int x, y, sx, sy;

    if (!g->activeplayer) return;

    x = ((c1.x > c2.x) ? c2.x : c1.x) * 20;
    if (x >= 20) x -= 20;
    y = ((c1.y > c2.y) ? c2.y : c1.y) * 20;
    if (y >= 20) y -= 20;
    sx = ((c1.x > c2.x) ? c1.x : c2.x)*20+40;
    sy = ((c1.y > c2.y) ? c1.y : c2.y)*20+40;

    canv->repaintContents(x, y, sx, sy, TRUE);
    }

void mainwndw::setwindow(coordinate c)
{   canv->ensureVisible(c.x * 20 + 10, c.y * 20 + 10);
    }

void mainwndw::invalidatemoves(void)
{   char str[50], rstr[8], hstr[30];

    if (selected) 
    {    sprintf(str, T_UNITMOVES, 
             selected->hitpointstring(hstr), selected->restmovestring(rstr));
         if (selected->is_transporter()) 
            if (((transporter *) selected)->getnextunload())
                 strcat(str, T_FULL); 
            else strcat(str, T_EMPTY);
         statusBar()->message(str);
         }
    else statusBar()->clear();
    }

void mainwndw::findnextunit(void)
{   class unit *current;
    int start = 0;

    current = (unit *) FIRSTMEMBER(g->units);
    if (!current) nextunit = (unit *) NULL;
    else
    {   if (selected)
	{   start = selected->getlfdnr();
	    current = (unit *) selected->next();
	    if (!current) current = (unit *) selected->first();
            }
	nextunit = current -> findnext(g->activeplayer, start);
        }
    if (nextunit) enable(nextbutton);
    else disable(nextbutton);
    }

void mainwndw::select(unit *u)
{   coordinate c;
    c.set(selectedpos.x, selectedpos.y);
    selectedpos.set(-1,-1);
    selected = u;
    invalidatesquare(c);
    if (selected && selected->restmoves()) enable(stoppbutton);
    else disable(stoppbutton);
    if (!selected) 
    {   selectedpos.set(0,0);
        invalidatemoves();
        disable(infobutton);
	//        festbutton->enable(0);
	return;
        }
    //    festbutton->enable(1);
    enable(infobutton);
    selected->unfortify();
    selectedpos = selected->getpos();

    g->unitmap[selectedpos] = selected->id();
    //    minebutton->enable((selected->id() == MINENLEGER)
    //        && !g->stadt(selectedpos) && !g->minebekannt(selectedpos));
    invalidatesquare(selectedpos);
    invalidatemoves();
    }

void mainwndw::selectcity(city *ct)
{   select((unit *) NULL);
    selectedpos = ct->getpos();
    setwindow(selectedpos);
    }

void mainwndw::nextplayer(void)
{   char msg[100];
    int i;
    char newgame;
    int inplay;
    active_socket *sock;

    if (waiting) return;
    uncity();
    unscore();
    select((unit *) NULL);

    newgame = !(g->activeplayer);

    i = g->get_next_player();

    g->activeplayer = 0;
    settitle();

    invalidate(FULL);

    if (g->sp[i].ext)
    {   waiting = 1;
        busy(1);
	if (!newgame)
	{   do
            {   sock = new active_socket(g->sp[i].host, g->sp[i].port);
                if (!(sock->active()))
                   QMessageBox::warning(this, "Emperor", T_ERRORNETSAVE, "OK");
                }
            while(!(sock->active()));

            g->activeplayer = i-1;

            g->save(sock->get_fd());  
            sock->close_me();
            delete sock;

	    }
	do
	    if (++i > 6) i = 1;
	while (!g->sp[i].used || g->sp[i].ext);
	g->activeplayer = i;
        settitle();
        listener = new listen_socket(g->get_port());
        qtm->start(NETWORK_POLL_INTERVAL);
	disable(nextbutton);
	// disable(minebutton);
	disable(stoppbutton);
	return;
        }
    inplay =  ((city *) FIRSTMEMBER(g->cities)) -> produce(i); // Production
    if (FIRSTMEMBER (g->units))
        inplay += ((unit *) FIRSTMEMBER(g->units)) -> clearmoves(i); // set move counter to 0

    g->lookout(i);
    if (forces) forces->update_armies();

    if (g->sp[i].computer)
    {   busy(1);
	g->activeplayer = i;
	g->compmove();
	busy(0);
	nextplayer();
        }
    else
    {   strcpy(msg, g->sp[i].name);
        if (g->get_turn() > 1)
            g->save("last.emp");

        if (!inplay) 
        {   strcat(msg, T_ELIMINATED);
            g->sp[i].used = 0;
            }
        else strcat(msg, T_YOURTURN);
        QMessageBox::information(this, "Message", msg);
	g->activeplayer = i;
	settitle();
        findnextunit();
        invalidate(MAP);
        button_next();
        }
   }

canvas::canvas(QWidget *parent, const char *name)
    : QScrollView(parent, name)
{   resizeContents(20*SX, 20*SY);
    landxpm = new QPixmap((const char **) land_xpm);
    flagsxpm = new QPixmap((const char **) flags);
    unitsxpm = new QPixmap((const char **) units);
    }

canvas::~canvas()
{   delete landxpm;
    delete flagsxpm;
    delete unitsxpm;
    }

void canvas::drawContents(QPainter *qp, int left, int top, int right, int bottom)
{   QColor grey("grey"), white("white"), black("black");
    QPen street(grey, 3);
    QPen whitePen(white, 2);

    int c, waiting;

    coordinate i, min, max, n, selectedpos;
    char p;
    class unit *selected;

    waiting = ((mainwndw *) parentWidget())->query_waiting();
    selected = ((mainwndw *) parentWidget())->query_selected();
    selectedpos = ((mainwndw *) parentWidget())->query_selectedpos();

    min.set(left - left % 20, top - top % 20);
    max.set(min.x + right + 20, min.y + bottom + 20);

    p = g->activeplayer;

    if (p && !waiting)
    {   for(i.x = min.x / 20; 20*i.x <= max.x ; i.x++)
            for(i.y = min.y / 20; 20*i.y <= max.y; i.y++)
            {   if (ONMAP(i))
		    if (g->bekannt(i))
                    {   if (g->strasse(i)) 
                        {   // display land square as background 
                            qp->drawPixmap(20*i.x, 20*i.y, *landxpm, 0, 60, 20, 20);
                            qp->setPen(street);
                            c = 1;
                            LOOP(n, -1, -1, 1, 1, 1) 
                               if (n.x || n.y)
                               {   if ((g->roadmap)[i] & c)
                                       qp->drawLine( 20*i.x+10, 20*i.y+10,
                                                    20*i.x+10 + n.x*10,
                                                    20*i.y+10 + n.y*10);
                                   c *= 2;
                                   }
                            }
                        else
                            qp->drawPixmap(20*i.x, 20*i.y, *landxpm, 
                                          0, 20*((*(g->mp))[i]), 20, 20);
                        if(g->minebekannt(i))
                        {    qp->fillRect(20*i.x+7, 20*i.y+7, 6, 6, white);
                             }
                        if(g->beobachtet(i) && g->ownership[i])
                             if (g->meins(i) || (g->unitmap[i] != UBOOT) || g->stadt(i))
                                  qp->drawPixmap(20*i.x+10, 20*i.y, *flagsxpm,
                                                0, 6*(g->ownership[i]-1), 10, 6);
                        if(g->beobachtet(i) && g->unitmap[i] 
                           && (g->meins(i) || (g->unitmap[i] != UBOOT)))
                              qp->drawPixmap(20*i.x, 20*i.y, *unitsxpm,
                                                0, 20*(g->unitmap[i]-1), 20, 20);
			 }
		    else
 		     	 qp->fillRect( 20*i.x, 20*i.y, 20, 20, black);
          	}
		       
        if (selected)
        {   qp->setPen(whitePen);
            qp->drawRect( (selectedpos.x)*20, (selectedpos.y)*20, 19, 19);
            }
        }
    else   // Busy... clear window 
        qp->fillRect( 0, 0, 20 * SX, 20 * SY, grey );
    }

void mainwndw::display_bird(void)
{   if (waiting) return;
    if (!bird)
        bird = new birdwndw(this);
    bird->show();
    }

void mainwndw::unbird(void)
{   if (!bird) return;
    bird->hide();
    }
  
void mainwndw::display_score(void)
{   if (waiting) return;
    if (!score)
        score = new scorewndw(this);
    score->show();
    }

void mainwndw::unscore(void)
{   if (!score) return;
    score->hide();
    }

void mainwndw::uncity(void)
{   if (!ctyw) return;
    ctyw->hide();
    }

void mainwndw::display_city(city *c)
{   if (!c) return;
    if (!ctyw)
        ctyw = new citywndw(this, c);
    else
        ctyw->setcity(c); // pass c to city
    ctyw->show();
    }

void mainwndw::display_armies(void)
{   if (waiting) return;
    if (!forces)
        forces = new armywndw(this);
    forces->show();
    }

void mainwndw::unforces(void)
{   if (!forces) return;
    forces->hide();
    }

int mainwndw::make_selection(coordinate c)
{   unit *u;

    if (!ONMAP(c)) return 0;
    if (!(g->bekannt(c))) return 0;
    u = (unit *) g->units.find(c);
    if (!u) return 0;
    if (u->getowner() != g->activeplayer) return 0;
    select(u);
    findnextunit();
    return 1;
    }

void mainwndw::make_move(coordinate moveto)
{   coordinate oldpos;
 
    if (!selected) return;
    if (!ONMAP(moveto)) return;
    if (selectedpos == moveto) return;

    oldpos = selectedpos;
    moveto = selected->movetowards(moveto);
    selected = selected->alive();

    if (selected)
    {   selectedpos = selected->getpos();
        if (!ONMAP(selectedpos)) selected = (unit *) NULL; 
        }
    select(selected);

    if (moveto.x == -1) moveto = oldpos; /* transporter load */
    invalidatearea(oldpos, moveto);
    invalidatemoves();
    findnextunit();
    if (forces) forces->update_armies();
    }

void canvas::contentsMousePressEvent(QMouseEvent *qme)
{   coordinate click;

    if (qme->button() != QMouseEvent::LeftButton) return;
 
    click.set(qme->x()/20, qme->y()/20);
    ((mainwndw *) parentWidget())->make_selection(click);
    }

void canvas::contentsMouseReleaseEvent(QMouseEvent *qme)
{   coordinate click;
    city *c;
    unit *selected;
    mainwndw *mw;

    mw = (mainwndw *) parentWidget();
    selected = mw->query_selected();
    

    click.set(qme->x()/20, qme->y()/20);    
    if (qme->button() == QMouseEvent::LeftButton)
    {   if (!selected) return;
        mw->make_move(click);
        return;
        }

    if (qme->button() == QMouseEvent::RightButton)
    {   if (!ONMAP(click)) return;
        if (!(g->bekannt(click))) return;

        /* Right click on a city: Open city window */
        c = (city *) g->cities.find(click);
        
        if (c) 
        {   mw->display_city(c);
            return;
            }

        /* Right click on a unit: Open info window */
        if (mw->make_selection(click))
        {   mw->button_info();
            return; 
            }

        /* Right click when a ferry/carrier is selected: Unload */
        if (selected && selected->is_transporter())
        {   ((transporter *) selected)->unload(click);
            mw->invalidatearea(click, click);
            mw->invalidatemoves(); 
            }
        }
    }

void mainwndw::keyPressEvent(QKeyEvent *qke)
{   coordinate click;
    int pace;

    pace = 20;
    if (qke->state() & ControlButton) pace *= 10;
    switch(qke->key())
    {   case Key_Left:
             canv->scrollBy(-pace, 0);
             return;
        case Key_Up:
             canv->scrollBy(0, -pace);
             return;
        case Key_Down:
             canv->scrollBy(0, pace);
             return;
        case Key_Right:
             canv->scrollBy(pace, 0);
             return;
        case Key_PageUp:
             canv->scrollBy(0, -200);
             return;
        case Key_PageDown:
             canv->scrollBy(0, 200);
             return;
        case Key_7:
             click.set(-1, -1);
             make_move(selectedpos + click);
             return;             
        case Key_8:
        case Key_K:
             click.set(0, -1);
             make_move(selectedpos + click);
             return;      
        case Key_9:
             click.set(1, -1);
             make_move(selectedpos + click);
             return;      
        case Key_4:
        case Key_H:
             click.set(-1, 0);
             make_move(selectedpos + click);
             return;     
        case Key_6:
        case Key_L:
             click.set(1, 0);
             make_move(selectedpos + click);
             return;          
        case Key_1:
             click.set(-1, 1);
             make_move(selectedpos + click);
             return;     
        case Key_2:
        case Key_J:
             click.set(0, 1);
             make_move(selectedpos + click);
             return;     
        case Key_3:
             click.set(1, 1);
             make_move(selectedpos + click);
             return;     
        case Key_F:
             button_stay();
             return;
        case Key_M:
             button_mine();
             return;
        case Key_R:
             button_road();
             return;
        case Key_Plus:
        case Key_Tab:
        case Key_5:
             button_next();
             return; 
        case Key_Space:
             button_stopp();
             return;
        }
 
    qke->ignore();
    }
/*
         case 'i':
              button_pressed(infobutton->get_handle());
              break;
     }
*/

void mainwndw::heartbeat()
{   active_socket *sock;

    /* check for incoming connections */
    sock = listener->accept_connection();

    if (sock)
    {  
        if (!(g->load(sock->get_fd()))) 
        {   QMessageBox::warning(this, "Emperor", T_ERRORNETLOAD, "OK");

            sock->close_me();
            delete sock;
            return;
            }        
 
        sock->close_me();
        delete sock;

        listener->close_me();
        delete listener;

        waiting = 0;
        busy(0);
        qtm->stop();
        nextplayer();
        }
    }

// **************************************************** citywndw ****

citywndw::citywndw(QWidget *parent, city *c) :
     QDialog(parent, (const char *) NULL)     
{   QPushButton *okbutton;
 
    setCaption(T_CITY);
    resize(160, 370);

    progress = new QProgressBar(this);
    progress->move(10, 10);
    progress->resize(140, 20);
                             
    auswahl = new QLabel(this);
    auswahl->move(10, 30); 

    unitbox = new QListBox(this);
    unitbox->move(10,90);
    unitbox->resize(140, 230);
    unitbox->insertItem(T_NOPRODUCTION, EXIT);
    unitbox->insertItem(T_ARMY, INFANTRIE);
    unitbox->insertItem(T_TANK, PANZER);
    unitbox->insertItem(T_AIRCRAFT, FLIEGER);
    unitbox->insertItem(T_SUBMARINE, UBOOT);
    unitbox->insertItem(T_BOAT, SCHNELLBOOT);
    unitbox->insertItem(T_FRIGATE, FREGATTE);
    unitbox->insertItem(T_BATTLESHIP, SCHLACHTSCHIFF);
    unitbox->insertItem(T_FERRY, TRANSPORTER);
    unitbox->insertItem(T_MINEBOAT, MINENLEGER);
    unitbox->insertItem(T_HELICOPTER, HELIKOPTER);
    unitbox->insertItem(T_BOMBER, BOMBER);
    unitbox->insertItem(T_CARRIER, FLUGZEUGTRAEGER);
    unitbox->hide();
    connect(unitbox, SIGNAL(selected(int)), this, SLOT(change_production(int)));
           
    bombing = new QLabel(this);
    bombing->move(10, 60);
    bombing->setAutoResize(TRUE);

    okbutton = new QPushButton(T_CLOSE, this);
    okbutton->move(30, 330);
    connect(okbutton, SIGNAL(clicked()), this, SLOT(hide()));
     
    setcity(c);
    show();
    }

citywndw::~citywndw(void)
{   delete unitbox;
    delete progress;
    }

void citywndw::setcity(city *c)
{   char bes[100];

    ct = c;

    strcpy(bes, T_CITY ": ");
    if (!g->beobachtet(ct->getpos()))
       strcat(bes, T_UNKNOWN);
    else
       strcat(bes, g->sp[ct->getowner()].name);
    setCaption(bes);

    if (ct->getowner() != g->activeplayer)
    {   unitbox->hide();
        progress->hide();
        auswahl->setText(g->sp[ct->getowner()].name);
        if (ct->get_bombs())
        {   bombing->show();
            bombing->setText(T_BOMBED);
            }
        else bombing->hide();
        return;
        }
    
    unitbox->show(); 

    if (ct->produced)
    {    progress->reset();
         progress->setTotalSteps(ct->produced->prodtime());
         progress->setProgress(ct->prodtime());
         progress->show();
         unitbox->setCurrentItem(ct->produced->id());
         auswahl->setText(ct->produced->name());
         }
    else 
    {    progress->hide();
         unitbox->setCurrentItem(0);
         auswahl->setText(T_NOPRODUCTION);
         }
    if (ct->get_bombs())
    {   bombing->show();
        sprintf(bes, T_TURNS_BOMBED, ct->get_bombs());
        bombing->setText(bes);
        }
    else bombing->hide();
    }

void citywndw::change_production(int i)
{   ct->setprod(i);
    setcity(ct);
    }

// *************************************************** scorewndw ****

scorewndw::scorewndw(QWidget * parent, const char *name) :
        QDialog(parent, name)
{   int i;
    QPixmap *qpx;
    
    resize(285, 220);
    setCaption(T_STATISTICS);
   
    citybn = new QPushButton(T_CITIES, this);
    citybn->move(0, 0);
    citybn->resize(95, 30);
    connect(citybn, SIGNAL(clicked()), this, SLOT(countcities()));

    areabn = new QPushButton(T_EXPLORED, this);
    areabn->move(95, 0);
    areabn->resize(95, 30);
    connect(areabn, SIGNAL(clicked()), this, SLOT(countarea()));

    armybn = new QPushButton(T_ARMIES, this);
    armybn->move(190, 0);
    armybn->resize(95, 30);
    connect(armybn, SIGNAL(clicked()), this, SLOT(countarmies()));

    for (i = 0; i < 6; i++)
    {   switch(i)
        {   case 0: qpx = new QPixmap((const char **) flag1); break;
            case 1: qpx = new QPixmap((const char **) flag2); break;
            case 2: qpx = new QPixmap((const char **) flag3); break;
            case 3: qpx = new QPixmap((const char **) flag4); break;
            case 4: qpx = new QPixmap((const char **) flag5); break;
            case 5: qpx = new QPixmap((const char **) flag6); break;
            }
        flagge[i] = new QLabel(this);
        flagge[i]->setPixmap(*qpx);
        flagge[i]->move(20, 40+30*i);             
//        delete qpx;
        score[i] = new QProgressBar(this);
        score[i]->move(55, 40+30*i);
        score[i]->resize(200, 20);
        } 

    countcities();
    show();
    }

scorewndw::~scorewndw(void)
{   int i;
    delete citybn;
    delete areabn;
    delete armybn;
    for (i = 0; i < 6; i++)  
    {   delete score[i];
        delete flagge[i];
        }
    }

void scorewndw::countcities(void)
{   city * run;
    int i, max;

    citybn->setEnabled(FALSE);
    areabn->setEnabled(TRUE);
    armybn->setEnabled(TRUE);

    for(i=1; i<7; i++) sc[i] = 0;
    max = 0;
    
    run = (city *) FIRSTMEMBER(g->cities);

    while(run)
    {   max++;
        sc[run->getowner()]++;
        run = (city *) run->next();
        }
    for(i=1; i<7; i++)
        if (sc[i])
        {    score[i-1]->reset();
             score[i-1]->setTotalSteps(max);
             score[i-1]->setProgress(sc[i]);
             score[i-1]->show();
             }
        else score[i-1]->hide();   
    setCaption(T_STATISTICS ": " T_CITIES);
    }

void scorewndw::countarea(void)
{   coordinate c;
    int i;

    citybn->setEnabled(TRUE);
    armybn->setEnabled(TRUE);
    areabn->setEnabled(FALSE);
    for(i=1; i<7; i++) sc[i] = 0;
    
    LOOP(c, 0, 0, SX-1, SY-1, 1)
        for(i=1; i<7; i++)
            if (g->bekannt(c, i)) sc[i]++;

    for(i=1; i<7; i++)
        if (sc[i])
        {    score[i-1]->reset();
             score[i-1]->setTotalSteps(100);
             sc[i] = ((long int) sc[i] * 100) / (SX * SY);
             score[i-1]->setProgress(sc[i]);
             score[i-1]->show();
             }
        else score[i-1]->hide();   
    setCaption(T_STATISTICS ": " T_EXPLORED);
    }


void scorewndw::countarmies(void)
{   unit *u;
    int i, max;

    citybn->setEnabled(TRUE);
    areabn->setEnabled(TRUE);
    armybn->setEnabled(FALSE);

    for(i=1; i<7; i++) sc[i] = 0;
    max = 0;
    
    u = (unit *) FIRSTMEMBER(g->units);

    while(u)
    {   max += u->gethp();
        sc[u->getowner()] += u->gethp();
        u = (unit *) u->next();
        }
    for(i=1; i<7; i++)
        if (sc[i])
        {    score[i-1]->reset();
             score[i-1]->setTotalSteps(100);
             sc[i] = ((long int) sc[i] * 100) / max;
             score[i-1]->setProgress(sc[i]);
             score[i-1]->show();
             }
        else score[i-1]->hide(); 
    setCaption(T_STATISTICS ": " T_ARMIES);
    }

// *************************************************** armywndw ****

armywndw::armywndw(QWidget * parent, const char *name) :
        QDialog(parent, name)
{   int i;
    char **unit_data;
    QPixmap *qpx;
    
    resize(95 + 40 * UNIT_TYPES, 225);
    setCaption(T_FORCES);
   
    for (i = 0; i < 6; i++)
    {   switch(i)
        {   case 0: qpx = new QPixmap((const char **) flag1); break;
            case 1: qpx = new QPixmap((const char **) flag2); break;
            case 2: qpx = new QPixmap((const char **) flag3); break;
            case 3: qpx = new QPixmap((const char **) flag4); break;
            case 4: qpx = new QPixmap((const char **) flag5); break;
            case 5: qpx = new QPixmap((const char **) flag6); break;
            }
        flagge[i] = new QLabel(this);
        flagge[i]->setPixmap(*qpx);
        flagge[i]->move(20, 40+30*i);
//        delete qpx;       
        }

    for (i = 1; i <= UNIT_TYPES; i++)
    {   unit_data = extract_unit(units, i);
        qpx = new QPixmap((const char **) unit_data);
        einheit[i-1] = new QLabel(this);
        einheit[i-1]->setPixmap(*qpx);
        free_array(unit_data);
        einheit[i-1]->move(60 + 40 * (i-1), 10);
//        delete qpx;
        }

    for(i=0; i<6*UNIT_TYPES; i++)
    {   zahl[i] = new QLabel("-", this);
        zahl[i]->move(60 + 40 * (i % UNIT_TYPES), 40 + 30 * (i / UNIT_TYPES));
        }

    unit_data = extract_land(land_xpm, 6);
    qpx = new QPixmap((const char **) unit_data);
    stadtpic = new QLabel(this);
    stadtpic->setPixmap(*qpx);
    free_array(unit_data);
    stadtpic->move(60 + (UNIT_TYPES * 40), 10);
//    delete qpx;

    for(i = 0; i < 6; i++)
    {   stadt[i] = new QLabel("-", this);
        stadt[i]->move(60 + (UNIT_TYPES * 40), 40 + 30 * i);
        }

    update_armies();
    show();
    }

armywndw::~armywndw(void)
{   int i;
    for (i = 0; i < 6; i++)  
       delete flagge[i];
    for (i = 0; i < UNIT_TYPES; i++)
       delete einheit[i];
    for (i = 0; i < 6 * UNIT_TYPES; i++)
       delete zahl[i]; 
    delete stadtpic;
    for (i = 0; i < 6; i++)
       delete stadt[i];
    }

void armywndw::update_armies(void)
{   int i;
    unit *u;
    city *run;
    char c[10];
    int sc[6 * UNIT_TYPES];

    for(i=0; i < 6*UNIT_TYPES; i++) sc[i] = 0;

    u = (unit *) FIRSTMEMBER(g->units);

    while(u)
    {   sc[(u->getowner()-1)* UNIT_TYPES + (u->id() -1)] ++;
        u = (unit *) u->next();
        }

    for(i = 0; i < 6*UNIT_TYPES; i++)
        if (!sc[i])
           zahl[i]->setText("-");
        else
        {  sprintf(c, "%d", sc[i]);
           zahl[i] -> setText(c);
           }

    for(i=1; i<7; i++) sc[i] = 0;
    
    run = (city *) FIRSTMEMBER(g->cities);

    while(run)
    {   sc[run->getowner()]++;
        run = (city *) run->next();
        }
    for(i=0; i<6; i++)
        if (!sc[i+1])
            stadt[i] -> setText("-");
        else
        {   sprintf(c, "%d", sc[i+1]);
            stadt[i] -> setText(c);
            }
    }
           
char ** armywndw::extract_unit(char ** pixdata, int unittype)
{   char **result;
    int i;

    result = (char **) malloc(24 * sizeof(char *));
    for(i = 0; i < 4; i++)
    {   result[i] = (char *) malloc(30 * sizeof(char));
        strcpy(result[i], pixdata[i]);
        }
    strcpy(result[0], "20 20 3 1"); 
    strcpy(result[1], "       c white");
    for(i = 0; i < 20; i++)
    {   result[4+i] = (char *) malloc(25 * sizeof(char));
        strcpy(result[4+i], pixdata[4 + 20 * (unittype -1) + i]);
        }
    return result;
    }

char ** armywndw::extract_land(char ** pixdata, int landtype)
{   char **result;
    int i;

    result = (char **) malloc(31 * sizeof(char *));
    for(i = 0; i < 11; i++)
    {   result[i] = (char *) malloc(30 * sizeof(char));
        strcpy(result[i], pixdata[i]);
        }
    strcpy(result[0], "20 20 10 1"); 
    for(i = 0; i < 20; i++)
    {   result[11+i] = (char *) malloc(25 * sizeof(char));
        strcpy(result[11+i], pixdata[11 + 20 * (landtype -1) + i]);
        }
    return result;
    }

void armywndw::free_array(char ** array)
{   int i;
    for (i = 0; i < 24; i++) free (array[i]);
    free (array);
    }    

// **************************************************** birdwndw ****

birdwndw::birdwndw(QWidget *parent, const char *name) :
        QDialog(parent, name)
{   resize(3*SX, 3*SY);
    setCaption(T_MAP);
    show();
    }

void birdwndw::invalidate(void)
{   update();
    }

void birdwndw::paintEvent(QPaintEvent *qpe)
{   QPainter qp(this); 
    QColor blue(0, 0, 192), lightblue(0, 192, 255);
    QColor green(0, 128, 0), red(255, 0, 0), black(0, 0, 0);
    QColor grey(192, 192, 192);
    coordinate i;

    if (g->activeplayer)
    {    LOOP(i, 0,0, SX-1, SY-1, 1)
             if (g->bekannt(i))
             {   switch((*(g->mp))[i])
                 {   case 0:
                     case 1:
                          qp.setBrush(blue);
                          break;
                     case 2:
                          qp.setBrush(lightblue);
                          break;
                     case 3:
                     case 4:
                          qp.setBrush(green);
                          break;
                     case 5:
                     case 6:
                          qp.setBrush(red);
                          break;
                     case 7:
                          qp.setBrush(black);
                          break;
                     }
                 qp.fillRect(3*i.x, 3*i.y, 3, 3, qp.brush());
                 }
         }
    else 
    {   qp.setBrush(grey);
        qp.fillRect(0, 0, 3*SX, 3*SY, qp.brush());
        }
    }

void birdwndw::mousePressEvent(QMouseEvent *qme)
{   coordinate click;
    click.set(qme->x()/3, qme->y()/3);
    ((mainwndw *) parentWidget())->setwindow(click);
    }
